#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import DocoptConfig
from inform import Error, plural
from stepwise import pl, ul, pre
from stepwise_mol_bio import Main

@autoprop
class SerialDilution(Main):
    """\
Perform a serial dilution.

Usage:
    serial_dilution <volume> <high> to <low> <steps> [options]
    serial_dilution <volume> <high> / <factor> <steps> [options]
    serial_dilution <volume> <low> x <factor> <steps> [options]

Arguments:
    <volume>
        The volume of reagent needed for each concentration.  A unit may be 
        optionally given, in which case it will be included in the protocol.

    <high>
        The starting concentration for the dilution.  A unit may be optionally 
        given, in which case it will be included in the protocol.

    <low>
        The ending concentration for the dilution.  A unit may be optionally 
        given, in which case it will be included in the protocol.

    <factor>
        How big of a dilution to make at each step of the protocol.

    <steps>
        The number of dilutions to make, including <high> and <low>.

Options:
    -m --material <name>    [default: ${app.material}]
        The substance being diluted.

    -d --diluent <name>     [default: ${app.diluent}]
        The substance to dilute into.

    -0 --include-zero
        Include a "dilution" with no material in the protocol.
"""
    __config__ = [
            DocoptConfig,
    ]

    _volume_str = appcli.param('<volume>')
    _conc_high_str = appcli.param('<high>', default=None)
    _conc_low_str = appcli.param('<low>', default=None)
    _factor = appcli.param('<factor>', cast=float, default=None)

    steps = appcli.param('<steps>', cast=int)
    material = appcli.param('--material', default='material')
    diluent = appcli.param('--diluent', default='water')
    include_zero = appcli.param('--include-zero', default=False)

    def __bareinit__(self):
        self._volume = None
        self._volume_unit = None

        self._conc_high = None
        self._conc_low = None
        self._conc_unit = None

    def __init__(self, volume, steps):
        self.volume = volume
        self.steps = steps

    def get_protocol(self):
        transfer = self.volume * (f := self.inv_factor) / (1 - f)
        initial_volume = self.volume + transfer

        conc_high_str = format_quantity(self.conc_high, self.conc_unit, 'g')
        material_str = f'{conc_high_str} {self.material}'.lstrip()
        conc_table = [
                [i, format_quantity(conc, self.conc_unit, 'e')]
                for i, conc in enumerate(self.concentrations, 1)
        ]

        num_tubes = self.steps if self.include_zero else self.steps - 1
        each_tube = 'each tube *except the last*' if self.include_zero else 'each tube'

        protocol = stepwise.Protocol()
        protocol += pl(
                "Perform a serial dilution [1]:",
                ul(
                    f"Put {initial_volume:.2f} μL {material_str} in a tube.",
                    f"Put {self.volume:.2f} μL {self.diluent} in {plural(num_tubes):# adjacent tube/s}.",
                    f"Transfer {transfer:.2f} μL between {each_tube} to make {self.steps} {self.factor:.2g}-fold dilutions.",
                ),
        )

        protocol.footnotes[1] = pl(
                "The final concentrations will be:",
                pre(stepwise.tabulate(conc_table, align='>>')),
                br='\n',
        )
        return protocol

    def get_concentrations(self):
        concs = [
                self.conc_high * self.factor**-i
                for i in range(self.steps)
        ]

        if self.include_zero:
            concs.append(0)

        return concs

    def get_volume(self):
        return self._volume

    def get_volume_unit(self):
        return self._volume_unit

    def set_volume(self, volume):
        self._volume, self._volume_unit = parse_quantity(volume)

    def get_conc_high(self):
        return self._conc_high

    def get_conc_low(self):
        return self._conc_low

    def get_conc_unit(self):
        return self._conc_unit

    def set_conc_high_low(self, high, low):
        self._conc_high, self._conc_low, self._conc_unit = parse_high_low(high, low)
        self._factor = (self._conc_high / self._conc_low)**(1 / (self.steps - 1))

    def set_conc_high_factor(self, high, factor):
        self._conc_high, self._conc_unit = parse_quantity(high)
        self._conc_low = self._conc_high * factor**-(self.steps - 1)
        self._factor = factor

    def set_conc_low_factor(self, low, factor):
        self._conc_low, self._conc_unit = parse_quantity(low)
        self._conc_high = self._conc_low * factor**(self.steps - 1)
        self._factor = factor

    def get_factor(self):
        return self._factor

    def get_inv_factor(self):
        return 1 / self._factor

    @appcli.on_load(DocoptConfig)
    def _load_volume(self):
        self.volume = self._volume_str

    @appcli.on_load(DocoptConfig)
    def _load_conc(self):
        if self._conc_high_str and self._conc_low_str:
            self.set_conc_high_low(
                    self._conc_high_str,
                    self._conc_low_str,
            )

        elif self._conc_high_str and self._factor:
            self.set_conc_high_factor(
                    self._conc_high_str,
                    self._factor,
            )

        elif self._conc_low_str and self._factor:
            self.set_conc_low_factor(
                    self._conc_low_str,
                    self._factor,
            )

def parse_quantity(x):
    try:
        return stepwise.Quantity.from_string(x).tuple
    except:
        return float(x), None

def parse_high_low(high_str, low_str):
    high, high_unit = parse_quantity(high_str)
    low, low_unit = parse_quantity(low_str)

    units = {high_unit, low_unit}
    units.discard(None)

    if len(units) > 1:
        raise Error(f"units don't match: {high_unit!r}, {low_unit!r}")

    return high, low, units.pop() if units else None

def format_quantity(value, unit=None, format='.2f', pad=' '):
    from numbers import Real
    if isinstance(value, Real):
        value = f'{value:{format}}'
    return f'{value}{pad}{unit}' if unit else value


if __name__ == '__main__':
    SerialDilution.main()

# vim: tw=53



#!/usr/bin/env python3

import appcli
import autoprop

from stepwise import Quantity
from stepwise_mol_bio import Main, UsageError
from freezerbox import (
        ReagentConfig, MakerConfig, mw_from_length, parse_conc, 
        parse_size_bp, parse_volume_uL, convert_conc_unit, iter_combos, 
        group_by_identity, group_by_cluster, unanimous, join_lists,
)
from appcli import Key, Method, DocoptConfig
from dataclasses import dataclass
from itertools import zip_longest
from more_itertools import one
from functools import partial
from collections import defaultdict
from inform import warn
from math import inf

# fragment:
#   Parameters for a single dsDNA construct, including name, concentration, 
#   molecular weight, volume, etc.
#
# assembly:
#   A list of fragments that will be assembled into a single plasmid.
#
# assemblies:
#   A list of assemblies that may be performed simultaneously, e.g. using 
#   master mixes.

# Right now all the assembly reactions used a fixed quantity of DNA, typically 
# 0.06 pmol.  I kinda think it might be better to just use presets, and to 
# scale the reaction volume normally.  But I'm not totally sure...

def parse_assemblies_from_docopt(args):
    assemblies = args['<assemblies>']
    concs = parse_fragment_attrs(
            args.get('--conc', []),
            partial(parse_conc, default_unit='ng/µL'),
    )
    lengths = parse_fragment_attrs(
            args.get('--length', []),
            partial(parse_size_bp, default_unit='bp'),
    )
    return parse_assemblies_from_comma_strs(
            args['<assemblies>'],
            concs=concs,
            lengths=lengths,
    )

def parse_assemblies_from_freezerbox(fields):
    return [parse_assembly_from_comma_str(fields[1])]

def parse_assemblies_from_comma_strs(assembly_strs, concs=None, lengths=None):
    return [
            parse_assembly_from_comma_str(x, concs=concs, lengths=lengths)
            for x in assembly_strs
    ]

def parse_assembly_from_comma_str(assembly_str, concs=None, lengths=None):
    if concs is None: concs = defaultdict(lambda: None)
    if lengths is None: lengths = defaultdict(lambda: None)

    assembly = []
    for name in assembly_str.split(','):
        fragment = Fragment(
                name,
                conc=concs[name],
                length=lengths[name],
        )
        assembly.append(fragment)

    return assembly

def parse_fragment_attrs(args, parse_value=lambda x: x):
    named_values = {}
    default_value = None

    for arg in args:
        fields = arg.split('=', 1)
        if len(fields) == 1:
            default_value = parse_value(arg)
        else:
            key, value = fields
            named_values[key] = parse_value(value)

    all_values = defaultdict(lambda: default_value)
    all_values.update(named_values)
    return all_values

def bind_assemblies(app, assemblies):
    for assembly in assemblies:
        for frag in assembly:
            frag.bind(app)
    return assemblies

def add_fragments_to_reaction(
        rxn,
        assemblies,
        *,
        target_pmol,
        min_pmol,
        excess_insert=1,
        order=-1,
    ):

    class FragmentReagent:

        def __init__(self, all_frags):
            self.fragments = all_frags

            # name, master_mix
            names = [x.name for x in all_frags]
            try:
                self.name = unanimous(names)
                self.master_mix = True
            except ValueError:
                self.name = ','.join(names)
                self.master_mix = False

            # conc
            frags = [x for x in all_frags if x is not null_reagent]
            self.conc_nM = inf
            self.conc = None

            for frag in frags:
                if frag.conc_nM < self.conc_nM:
                    self.conc_nM = frag.conc_nM
                    self.conc = frag.conc

            def all_close(x, threshold=0.1):
                try:
                    a, b = max(x), min(x)
                    return (a - b) / a < 0.1
                except TypeError:
                    return False

            concs = [x.conc for x in frags]
            concs_nM = [x.conc_nM for x in frags]

            if not all_close(concs_nM) or not all_close(concs):
                self.conc = self.conc_nM, 'nM'

            # volume
            self.vol_uL = uL_from_pmol(target_pmol, self.conc_nM)

    null_reagent = Fragment('−')
    frag_reagents = [
            FragmentReagent(frags)
            for frags in zip_longest(*assemblies, fillvalue=null_reagent)
    ]
    for i, reagent in enumerate(frag_reagents):
        excess = 1 if i == 0 else excess_insert
        reagent.vol_uL *= excess

    # Make sure the maximum volume is not exceeded.
    total_vol_uL = sum(x.vol_uL for x in frag_reagents)
    max_vol_uL = rxn.free_volume.value 
    if max_vol_uL < total_vol_uL:
        k = max_vol_uL / total_vol_uL
        for reagent in frag_reagents:
            reagent.vol_uL *= k

    # Warn if any fragment is below the minimum.
    for reagent in frag_reagents:
        reagent.pmol = pmol_from_uL(reagent.vol_uL, reagent.conc_nM)
        if reagent.pmol < min_pmol:
            warn(f"using {reagent.pmol:.3f} pmol of {reagent.name}, {min_pmol:.3f} pmol recommended.")

    for reagent in frag_reagents:
        rxn[reagent.name].volume = reagent.vol_uL, 'µL'
        rxn[reagent.name].stock_conc = reagent.conc
        rxn[reagent.name].master_mix = reagent.master_mix
        rxn[reagent.name].order = order
        rxn[reagent.name].flags.add('fragment')
        rxn[reagent.name].pmol = reagent.pmol

    rxn.num_reactions = len(assemblies)

def uL_from_pmol(pmol, conc_nM):
    return 1e3 * pmol / conc_nM

def pmol_from_uL(uL, conc_nM):
    return uL * conc_nM / 1e3

ARGUMENT_DOC = """\
    <assemblies>...
        The names of the DNA fragments to assemble, separated by commas.  Each 
        argument defines a single assembly reaction.  You can specify any 
        number of arguments to setup multiple assemblies at once.  Any 
        fragments that are used in every assembly will be automatically 
        included in the master mix.

        The concentration of every fragment must be known.  If the fragment 
        name corresponds to an entry in the FreezerBox database, the 
        concentration specified in the database will be used by default.  
        Otherwise, you must specify concentrations using the `--conc` option.

        The first fragment in each assembly is considered the "backbone", and 
        all the remaining fragments are considered "inserts".  This is only 
        relevant for the `--excess-insert` option."""
OPTION_DOC = """\
    -v --volume <µL>                [default: ${app.volume_uL}]
        The volume of the complete assembly reaction.  This does not affect the 
        amount of DNA that will be used; the recomendations for that amount are 
        a fixed number of pmol per reaction.  But it may be necessary to 
        increase the volume if your DNA is dilute, or if you have a large 
        number of inserts.

    -c --conc <[name=]value>
        Either the concentration of a single fragment, or the default 
        concentration for all fragments.  The former applies if a name is 
        specified (followed by an equals sign); the latter applies 
        otherwise.

        The concentration value can optionally include a unit, e.g. nM, ng/µL, 
        etc.  If no unit is specified, ng/µL is assumed.  If the concentration 
        is not specified in terms of molarity, the length of the fragment(s) in 
        question must also be specified.  This information can be queried from 
        the FreezerBox database, or explicitly given via the `--length` option.

        Although concentrations can be read from the FreezerBox database, any 
        values specified on the command-line will override any specified in the 
        database.

    -l --length <[name=]value>
        Either the length of a single fragment, or the default length for all 
        fragments.  The former applies if a name is specified (followed by an 
        equals sign); the latter applies otherwise.

        The length can optionally include a unit, e.g. bp or kb.  If no unit is 
        specified, bp is assumed.

        Lengths only need to be specified for fragments that don't have 
        concentrations in molarity units.  The lengths will be used to estimate 
        to molecular weight of each fragment.  This assumes that each fragment 
        is dsDNA.  For fragments that are present in the FreezerBox database, 
        molecular wieght will be inferred by default from any sequence or 
        length information present in the database.  Lengths specified on the 
        command-line override any values present in the database.

    -x --excess-insert <ratio>      [default: ${app.excess_insert}]
        The molar-excess of each insert relative to the backbone.  Values 
        between 1-10 (e.g. 1-10x excess) are typical."""

@autoprop.cache
class Fragment:
    __config__ = [ReagentConfig]

    name = tag = appcli.param()
    conc = appcli.param(
            Key(ReagentConfig),
    )
    mw = appcli.param(
            Key(ReagentConfig),
            Method(lambda self: mw_from_length(self.length)),
            default=None,
    )
    length = appcli.param(
            Key(ReagentConfig),
    )

    def __init__(self, name=None, *, conc=None, length=None, mw=None):
        if name: self.name = name
        if conc: self.conc = conc
        if length: self.length = length
        if mw: self.mw = mw

    def __repr__(self):
        attrs = 'name', 'conc', 'length', 'mw'
        attr_strs = [
                f'{attr}={value!r}'
                for attr in attrs
                if (value := getattr(self, attr, None)) is not None
        ]
        return f'Fragment({", ".join(attr_strs)})'

    def __eq__(self, other):
        undef = object()
        attrs = 'name', 'conc', 'length'
        return all(
                getattr(self, attr, undef) == getattr(other, attr, undef)
                for attr in attrs
        )

    def bind(self, app, force=False):
        if not hasattr(self, 'app') or force:
            self.app = app

    def get_db(self):
        return self.app.db

    def get_conc_nM(self):
        return convert_conc_unit(self.conc, self.mw, 'nM').value

    def del_conc_nM(self):
        pass

@autoprop
class Assembly(Main):
    __config__ = [
            DocoptConfig,
            MakerConfig,
    ]

    Fragment = Fragment
    target_pmol_per_frag = 0.06
    min_pmol_per_frag = 0.02

    assemblies = appcli.param(
            Key(DocoptConfig, parse_assemblies_from_docopt),
            Key(MakerConfig, parse_assemblies_from_freezerbox),
            get=bind_assemblies
    )
    volume_uL = appcli.param(
            Key(DocoptConfig, '--volume', cast=float),
            Key(MakerConfig, 'volume', cast=parse_volume_uL),
            default=5,
    )
    excess_insert = appcli.param(
            Key(DocoptConfig, '--excess-insert', cast=float),
            default=2,
    )

    group_by = {
            'volume_uL': group_by_identity,
            'excess_insert': group_by_identity,
    }
    merge_by = {
            'assemblies': join_lists,
    }

    def __init__(self, assemblies):
        self.assemblies = assemblies

    def get_num_fragments(self):
        return max(len(x) for x in self.assemblies)

    def get_product_conc(self):
        min_pmol = min(
                x.pmol
                for x in self.reaction.iter_reagents_by_flag('fragment')
        )
        return Quantity(1e3 * min_pmol / self.volume_uL, 'nM')

    def get_dependencies(self):
        return {
                frag.name
                for assembly in self.assemblies
                for frag in assembly
        }

    def _add_fragments_to_reaction(self, rxn, order=-1):
        rxn.hold_ratios.volume = self.volume_uL, 'µL'
        rxn.extra_min_volume = '0.5 µL'
        add_fragments_to_reaction(
                rxn, self.assemblies,
                target_pmol=self.target_pmol_per_frag,
                min_pmol=self.min_pmol_per_frag,
                excess_insert=self.excess_insert,
                order=order,
        )
        return rxn




#!/usr/bin/env python3

import pytest
import freezerbox

@pytest.fixture(autouse=True)
def ignore_external_freezerbox_db(monkeypatch):
    mock_db = freezerbox.Database("WARNING: ACCESSING EXTERNAL DATABASE")
    mock_db_factory = lambda: mock_db

    monkeypatch.setattr(freezerbox, 'load_db', mock_db_factory)
    monkeypatch.setattr(freezerbox.model, 'load_db', mock_db_factory)




#!/usr/bin/env python3

import stepwise, appcli, autoprop, freezerbox
from math import sqrt, ceil
from numbers import Real
from inform import plural, indent
from appcli import Key, Method, DocoptConfig
from stepwise import (
        StepwiseConfig, PresetConfig, Quantity, UsageError,
        pl, ul, pre
)
from stepwise_mol_bio import (
        Main, Argument, ShareConfigs, ConfigError,
        comma_set, bind_arguments, require_reagent,
        int_or_expr, float_or_expr, merge_names,
)
from freezerbox import (
        ReagentConfig, MakerConfig, unanimous,
        parse_volume_uL, parse_temp_C, parse_time_s,
        group_by_identity, join_lists,
)
from more_itertools import first_true, flatten, chunked, all_equal, always_iterable
from statistics import mean
from collections.abc import Iterable
from copy import deepcopy
from operator import not_

# I'd like to support multiple <product> arguments, but first I need 
# MakerConfig to support multiple products (e.g. call unanimous, then fail 
# gracefully).

def find_amplicon(template, primer_1, primer_2, is_template_circular=False):
    from Bio.Seq import reverse_complement

    # This is a temporary solution until I implement full support for IDT-style 
    # sequence strings in freezerbox.
    template = freezerbox.normalize_seq(template)
    primer_1 = freezerbox.normalize_seq(primer_1)
    primer_2 = freezerbox.normalize_seq(primer_2)

    primer_pairs = [
            (primer_1, reverse_complement(primer_2)),
            (primer_2, reverse_complement(primer_1)),
    ]

    for fwd, rev in primer_pairs:
        # Assume perfect complementarity in the last 15 bases.  This isn't a 
        # good approach; better would be to predict where each oligo would 
        # anneal.  I think I could use Smith-Waterman for this.
        i = template.find(fwd[-15:  ])
        j = template.find(rev[   :15])

        if i < 0 or j < 0:
            continue

        if i < j:
            break

        if i > j:
            if is_template_circular:
                break
            else:
                continue

        if i == j:
            raise ValueError("primers bind same position")

    else:
        raise ValueError("no amplicon found")

    if i < j:
        return fwd[:-15] + template[i:j] + rev
    else:
        return fwd[:-15] + template[i:] + template[:j] + rev

def parse_amplicons_from_docopt(args):
    return [parse_amplicon(x) for x in args['<amplicons>']]

def parse_amplicons_from_freezerbox(fields):
    template = fields['template']
    fwd, rev = parse_primers(fields['primers'])
    return [Pcr.Amplicon.from_tags(template, fwd, rev)]

def parse_amplicon(string):
    fields = string.split(',')
    n = len(fields) if string else 0

    if n == 3:
        return Pcr.Amplicon.from_tags(*fields)

    err = ConfigError(string=string, n=n)
    err.brief = "can't interpret {string!r} as an amplicon"
    err.info += "expected 3 comma-separated fields (template, fwd, rev)"
    err.blame += "found {n}"
    raise err

def parse_primers(string):
    fields = string.split(',')
    n = len(fields) if string else 0

    if n == 2:
        return fields

    err = ConfigError(string=string, n=n)
    err.brief = "can't interpret {string!r} as a primer pair"
    err.info += "expected 2 comma-separated fields (fwd, rev)"
    err.blame += "found {n}"
    raise err

def temp(x):
    try:
        return float(x)
    except ValueError:
        return x

def time(x):
    return float(x)

@autoprop.cache
class Pcr(Main):
    """\
Amplify a DNA template using polymerase chain reaction (PCR).

Usage:
    pcr <amplicon>... [-a <°C>] [-x <sec> | -l <kb>] [options]
    pcr (-u <product>) [options]

Arguments:
    <amplicon>
        The names of the templates and forward/reverse primers to use for each 
        reaction, separated by commas.  If a `freezerbox` database is present 
        and these names can be found in it, default values for a number of 
        parameters (e.g. annealing temperate, extension time, etc.) will be 
        derived from it.

<%! from stepwise_mol_bio import hanging_indent %>\
Options:
    -p --preset <name>              [default: ${app.preset}]
        The default reaction and thermocycler parameters to use.  The following 
        sets of parameters are currently available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    -u --product <name>
        The name of a product to make.  This name must be present in the 
        FreezerBox database.  In this form of the command, all default settings 
        will be taken from the reaction used to synthesize the given product.

    -l --amplicon-length <bp>
        The length of the amplicon in base pairs (bp).  This can be used to 
        calculate an appropriate extension time.

    -n --num-reactions <int>        [default: ${app.num_reactions}]
        The number of reactions to set up.

    -v --reaction-volume <μL>
        The volume of the PCR reaction.

    -V --template-volume <µL>
        The volume of template to use in the reaction.  This overrides the 
        value specified by the preset.

    -T --template-stock <conc>
        The stock concentration of template to use in the reaction.  This 
        overrides the value specified by the preset, without affecting the 
        volume.  Include a unit, because none is implied.

    --primer-stock <µM>
        The stock concentration of the primers in µM.  Both primers must have 
        the same concentration.

    -y --num-cycles <n>
        The number of denature/anneal/extend cycles to perform, e.g. 35.

    --initial-denature-temp <°C>
        The temperature of the initial denaturation step in °C, e.g. 95.

    --initial-denature-time <sec>
        The duration of the initial denaturation step in seconds, e.g. 30.

    --denature-temp <°C>
        The temperature of the denaturation step in °C, e.g. 95.

    --denature-time <sec>
        The duration of the denaturation step in seconds, e.g. 10.

    -a --anneal-temp <°C>
        The temperature of the annealing step in °C, e.g. 60.  This is 
        determined by the sequence of the primers.
        
    -g --anneal-temp-gradient <range>
        The range of annealing temperatures that should be tried in a gradient 
        PCR reaction.  The range will be centered at the indicated annealing 
        temperature, and the protocol will indicate the corresponding high and 
        low temperatures.

    --anneal-time <sec>
        The duration of the annealing step in seconds, e.g. 20.

    --extend-temp <°C>
        The temperature of the extension step in °C, e.g. 72.
        
    -x --extend-time <sec>
        The duration of the extension step in seconds.  The rule of thumb is
        30 sec/kb, perhaps longer if you're amplifying a whole plasmid.

    --no-round-extend-time 
        When calculating an extension time from a given amplicon length (e.g. 
        `-l`), do *not* round the result to the nearest 30 sec increment.  Note 
        that when an explicit extension time is given (e.g. `-x`), it is never 
        rounded.

    --final-extend-temp <°C>
        The temperature of the final extension step in °C, e.g. 72.

    --final-extend-time <sec>
        The duration of the annealing step in seconds, e.g. 120.

    --hold-temp <°C>
        The temperature in °C to hold the reaction at after it completes.

    --melt-curve-low-temp <°C>
        The temperature in °C at which to begin recording a melt curve,
        e.g. 65.  This is only relevant for qPCR protocols.

    --melt-curve-high-temp <°C>
        The temperature in °C at which to stop recording a melt curve,
        e.g. 95.  This is only relevant for qPCR protocols.

    --melt-curve-temp-step <°C>
        How much to increment the temperature in °C at each step in the melt 
        curve, e.g. 0.5°C.  This is only relevant for qPCR protocols.

    --melt-curve-time-step <sec>
        The duration in seconds of each step in the melt curve, e.g. 5.  This 
        is only relevant for qPCR protocols.

    --two-step
        Specify that the annealing and extension steps should be combined, e.g. 
        if the primer melting temperatures are very close to the extension 
        temperature or if the amplicon is very short.

    --qpcr
        Specify that this is a qPCR protocol, i.e. that fluorescence should be 
        measured after each thermocyler cycle.
"""

    __config__ = [
            DocoptConfig.setup(usage_getter=lambda self: self.format_usage()),
            MakerConfig,
            PresetConfig,
            StepwiseConfig.setup('molbio.pcr'),
    ]
    usage = appcli.config_attr()
    preset_briefs = appcli.config_attr()

    @autoprop
    class Amplicon(ShareConfigs, Argument):
        __config__ = []

        def _calc_seq(self):
            try:
                return find_amplicon(
                        self.template.seq,
                        self.fwd.seq,
                        self.rev.seq,
                        self.template.is_circular,
                )
            except ValueError:
                err = ConfigError(
                        template=self.template,
                        fwd=self.fwd,
                        rev=self.rev,
                )
                err.brief = "{fwd.tag!r} and {rev.tag!r} do not amplify {template.tag!r}"
                err.info += "{template.tag}: {template.seq}"
                err.info += "{fwd.tag}: {fwd.seq}"
                err.info += "{rev.tag}: {rev.seq}"
                raise err from None

        seq = appcli.param(
                Method(_calc_seq),
        )
        length_bp = appcli.param(
                Key(DocoptConfig, '--amplicon-length', cast=float),
                Method(lambda self: len(self.seq)),
                default=None
        )

        @classmethod
        def from_tags(cls, template, fwd, rev, **kwargs):
            return cls(
                    Pcr.Template(template),
                    Pcr.Primer(fwd),
                    Pcr.Primer(rev),
                    **kwargs,
            )

        def __init__(self, template, fwd, rev, **kwargs):
            self.template = template
            self.fwd = fwd
            self.rev = rev
            self._set_known_attrs(kwargs)

        def __repr__(self):
            return f'{self.__class__.__qualname__}({self.template!r}, {self.fwd!r}, {self.rev!r})'

        def get_reagents(self):
            return self.template, *self.primers

        def get_reagent_tags(self):
            return tuple(x.tag for x in self.reagents)

        def get_primers(self):
            return self.fwd, self.rev

        def set_primers(self, primers):
            self.fwd, self.rev = primers

        def on_bind(self, app):
            super().on_bind(app)
            for reagent in self.reagents:
                reagent.bind(app)

    class Template(Argument):
        __config__ = [ReagentConfig]

        seq = appcli.param('seq')
        is_circular = appcli.param('is_circular')

    class Primer(ShareConfigs, Argument):
        __config__ = [ReagentConfig]

        seq = appcli.param(
                Key(ReagentConfig, 'seq'),
        )
        melting_temp_C = appcli.param(
                Key(ReagentConfig, 'melting_temp_C'),
        )
        stock_uM = appcli.param(
                Key(DocoptConfig, '--primer-stock', cast=float),
                Key(ReagentConfig, 'conc_uM'),
                Key(PresetConfig, 'primer_stock_uM'),
                Key(StepwiseConfig, 'primer_stock_uM'),
        )

    def _calc_anneal_temp_C(self):
        return [
                self.anneal_temp_func(
                    amplicon.fwd.melting_temp_C,
                    amplicon.rev.melting_temp_C,
                )
                for amplicon in self.amplicons
        ]

    def _calc_extend_time_s(self):
        bp = min(bp for x in self.amplicons if (bp := x.length_bp))

        def round(x):
            if not self.round_extend_time: return x
            if x <= 10: return 10
            if x <= 15: return 15
            return 30 * ceil(x / 30)

        if factor := self.extend_time_s_per_kb:
            return round(factor * bp / 1000)

        return round(self.extend_time_func(bp))

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=list,
    )
    preset = appcli.param(
            Key(DocoptConfig, '--preset'),
            Key(MakerConfig, 'preset'),
            Key(StepwiseConfig, 'default_preset'),
    )
    amplicons = appcli.param(
            Key(DocoptConfig, parse_amplicons_from_docopt),
            Key(MakerConfig, parse_amplicons_from_freezerbox),
            get=bind_arguments,
    )
    product_tag = appcli.param(
            Key(DocoptConfig, '--product'),
    )
    products = appcli.param(
            Method(lambda self: [
                self.db[self.product_tag].make_intermediate(0)
            ]),
    )
    template_volume_uL = appcli.param(
            Key(DocoptConfig, '--template-volume', cast=float),
            default=None,
    )
    template_stock = appcli.param(
            Key(DocoptConfig, '--template-stock'),
            # Don't try to get a value from the amplicons/freezerbox database, 
            # because the template volume is not affected by this setting.
            default=None,
    )
    product_conc_ng_uL = appcli.param(
            Key(PresetConfig, 'product_conc_ng_uL'),
            default=50,
    )
    num_reactions = appcli.param(
            Key(DocoptConfig, '--num-reactions', cast=int_or_expr),
            Method(lambda self: len(self.amplicons)),
    )
    reaction_volume_uL = appcli.param(
            Key(DocoptConfig, '--reaction-volume'),
            Key(MakerConfig, 'volume', cast=parse_volume_uL),
            Key(PresetConfig, 'reaction_volume_uL'),
            cast=float_or_expr,
            default=None,
    )
    extra_volume_uL = appcli.param(
            Key(PresetConfig, 'extra_volume_uL'),
            Key(StepwiseConfig, 'extra_volume_uL'),
            cast=float_or_expr,
    )
    extra_percent = appcli.param(
            Key(PresetConfig, 'extra_percent'),
            Key(StepwiseConfig, 'extra_percent'),
            cast=float_or_expr,
    )
    base_reaction = appcli.param(
            Key(PresetConfig, 'reagents'),
            cast=stepwise.MasterMix.from_text,
    )
    num_cycles = appcli.param(
            Key(DocoptConfig, '--num-cycles'),
            Key(PresetConfig, 'num_cycles'),
            cast=int,
    )
    initial_denature_temp_C = appcli.param(
            Key(DocoptConfig, '--initial-denature-temp'),
            Key(PresetConfig, 'initial_denature_temp_C'),
            cast=temp,
    )
    initial_denature_time_s = appcli.param(
            Key(DocoptConfig, '--initial-denature-time'),
            Key(PresetConfig, 'initial_denature_time_s'),
            cast=time,
    )
    denature_temp_C = appcli.param(
            Key(DocoptConfig, '--denature-temp'),
            Key(PresetConfig, 'denature_temp_C'),
            cast=temp,
    )
    denature_time_s = appcli.param(
            Key(DocoptConfig, '--denature-time'),
            Key(PresetConfig, 'denature_time_s'),
            cast=time,
    )
    anneal_temp_C = appcli.param(
            Key(DocoptConfig, '--anneal-temp', cast=temp),
            Key(MakerConfig, 'Ta', cast=parse_temp_C),
            Method(_calc_anneal_temp_C),
            Key(PresetConfig, 'anneal_temp_C'),
    )
    anneal_temp_func = appcli.param(
            Key(PresetConfig, 'anneal_temp_func', cast=eval),
            default=lambda x: min(x) + 1,
    )
    anneal_temp_gradient_C = appcli.param(
            Key(DocoptConfig, '--anneal-temp-gradient'),
            Key(PresetConfig, 'anneal_temp_gradient_C'),
            cast=float,
    )
    anneal_time_s = appcli.param(
            Key(DocoptConfig, '--anneal-time'),
            Key(PresetConfig, 'anneal_time_s'),
            cast=time,
    )
    extend_temp_C = appcli.param(
            Key(DocoptConfig, '--extend-temp'),
            Key(PresetConfig, 'extend_temp_C'),
            cast=temp,
    )
    extend_time_s = appcli.param(
            Key(DocoptConfig, '--extend-time', cast=time),
            Key(MakerConfig, 'tx', cast=parse_time_s),
            Method(_calc_extend_time_s, skip=ValueError),
            Key(PresetConfig, 'extend_time_s'),
    )
    extend_time_s_per_kb = appcli.param(
            Key(PresetConfig, 'extend_time_s_per_kb'),
            cast=float,
    )
    extend_time_func = appcli.param(
            Key(PresetConfig, 'extend_time_func', cast=eval),
    )
    round_extend_time = appcli.param(
            Key(DocoptConfig, '--no-round-extend-time', cast=not_),
            Key(PresetConfig, 'round_extend_time'),
            Key(StepwiseConfig, 'round_extend_time'),
            default=True,
    )
    final_extend_temp_C = appcli.param(
            Key(DocoptConfig, '--final-extend-temp'),
            Key(PresetConfig, 'final_extend_temp_C'),
            cast=temp,
    )
    final_extend_time_s = appcli.param(
            Key(DocoptConfig, '--final-extend-time'),
            Key(PresetConfig, 'final_extend_time_s'),
            cast=time,
    )
    hold_temp_C = appcli.param(
            Key(DocoptConfig, '--hold-temp'),
            Key(PresetConfig, 'hold_temp_C'),
            cast=temp,
            default=None,
    )
    melt_curve_low_temp_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-low-temp'),
            Key(PresetConfig, 'melt_curve_low_temp_C'),
            cast=temp,
    )
    melt_curve_high_temp_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-high-temp'),
            Key(PresetConfig, 'melt_curve_high_temp_C'),
            cast=temp,
    )
    melt_curve_temp_step_C = appcli.param(
            Key(DocoptConfig, '--melt-curve-temp-step'),
            Key(PresetConfig, 'melt_curve_temp_step_C'),
            cast=temp,
    )
    melt_curve_time_step_s = appcli.param(
            Key(DocoptConfig, '--melt-curve-time-step'),
            Key(PresetConfig, 'melt_curve_time_step_s'),
            cast=temp,
    )
    two_step = appcli.param(
            Key(DocoptConfig, '--two-step'),
            Key(PresetConfig, 'two_step'),
            pick=first_true,
            default=False,
    )
    qpcr = appcli.param(
            Key(DocoptConfig, '--qpcr'),
            Key(PresetConfig, 'qpcr'),
            pick=first_true,
            default=False,
    )
    footnote = appcli.param(
            Key(PresetConfig, 'footnote'),
            default=None,
    )

    group_by = {
        'preset': group_by_identity,
        'reaction_volume_uL': group_by_identity,
    }
    merge_by = {
        'amplicons': join_lists,
        'anneal_temp_C': list,
        'extend_time_s': max,
    }

    @classmethod
    def from_product(cls, product_tag):
        self = cls.from_params()
        self.product_tag = product_tag
        self.load(MakerConfig)
        return self

    def __init__(self, amplicons):
        self.amplicons = amplicons

    def format_usage(self):
        return self.__doc__

    def get_protocol(self):
        protocol = stepwise.Protocol()
        pcr, primer_mix = self.reaction
        thermocycler = self.thermocycler_protocol
        footnotes = []

        # Primer mix (if applicable):

        example_uM = min(x.stock_uM for x in flatten(self.primer_pairs))
        footnotes.append(pl(
                f"For resuspending lyophilized primers:",
                f"{example_uM} µM = {1e3 / example_uM:g} µL/nmol",
                br='\n',
        ))

        if primer_mix:
            protocol += pl(
                    f"Prepare 10x primer mix{protocol.add_footnotes(*footnotes)}:",
                    primer_mix,
            )
            footnotes = []

        # PCR reaction setup:

        if self.footnote:
            # Before the footnote on resuspending the primers, if it hasn't 
            # been added to the protocol already.
            footnotes.insert(0, self.footnote)

        if x := pcr['template DNA'].stock_conc:
            footnotes.append(pl(
                    f"For diluting template DNA to {x}:",
                    f"Dilute 1 µL twice into {sqrt(1000/x.value):.1g}*sqrt([DNA]) µL",
                    br='\n',
            ))

        title = 'qPCR' if self.qpcr else 'PCR'
        instructions = ul()
        protocol += pl(
                f"Setup {plural(pcr.num_reactions):# {title} reaction/s}{protocol.add_footnotes(*footnotes)}:",
                pcr,
                instructions,
        )

        if pcr.volume > '50 µL':
            instructions += f"Split each reaction into {ceil(pcr.volume.value / 50)} tubes."
        if pcr.num_reactions > 1:
            instructions += "Use any extra master mix as a negative control."

        # Thermocycler protocol:

        protocol += pl(
                "Run the following thermocycler protocol:",
                thermocycler,
        )

        protocol.renumber_footnotes()
        return protocol

    def get_thermocycler_protocol(self):

        def temp(x):
            if isinstance(x, Iterable) and not isinstance(x, str):
                return merge_names(map(temp, x))
            elif isinstance(x, Real):
                return f'{x:g}°C'
            elif x[-1].isdigit():
                return f'{x}°C'
            else:
                return x

        def time(x):
            if isinstance(x, Iterable):
                return ','.join(map(temp, x))
            elif isinstance(x, str):
                return x
            elif x < 60:
                return f'{x:.0f}s'
            elif x % 60:
                return f'{x//60:.0f}m{x%60:02.0f}'
            else:
                return f'{x//60:.0f} min'

        def has_step(step, params=['temp_C', 'time_s']):
            return all(
                hasattr(self, f'{step}_{param}')
                for param in params
            )

        def step(step):
            if not hasattr(self, f'{step}_temp_gradient_C'):
                t = getattr(self, f'{step}_temp_C')
            else:
                t_mid = getattr(self, f'{step}_temp_C')
                t_mid = mean(always_iterable(t_mid))
                t_range = getattr(self, f'{step}_temp_gradient_C')
                t_low = round(t_mid - t_range / 2)
                t_high = round(t_low + t_range)
                t = f'{t_low}-{t_high}'

            return f"{temp(t)} for {time(getattr(self, f'{step}_time_s'))}"

        three_step = not self.two_step and has_step('extend')

        thermocycler_steps = [
                f"- {step('initial_denature')}",
                f"- Repeat {getattr(self, 'num_cycles')}x:",
                f"  - {step('denature')}",
                f"  - {step('anneal')}",
        ]
        if three_step:
            thermocycler_steps += [
                f"  - {step('extend')}",
            ]

        if self.qpcr:
            thermocycler_steps += [
                f"  - Measure fluorescence",
            ]

        if has_step('final_extend'):
            thermocycler_steps += [
                f"- {step('final_extend')}",
            ]

        if has_step('melt_curve', 'low_temp_C high_temp_C temp_step_C time_step_s'.split()):
            thermocycler_steps += [
                f"- {self.melt_curve_low_temp_C}-{self.melt_curve_high_temp_C}°C in {time(self.melt_curve_time_step_s)} steps of {self.melt_curve_temp_step_C}°C:",
            ]

        if self.qpcr:
            thermocycler_steps += [
                f"  - Measure fluorescence",
            ]

        if self.hold_temp_C:
            thermocycler_steps += [
                f"- {temp(self.hold_temp_C)} hold",
            ]

        return pre('\n'.join(thermocycler_steps))

    def get_reaction(self):
        pcr = deepcopy(self.base_reaction)
    
        require_reagent(pcr, 'water')
        require_reagent(pcr, 'template DNA')
        require_reagent(pcr, 'forward primer')
        require_reagent(pcr, 'reverse primer')

        pcr.extra_volume = self.extra_volume_uL, 'µL'
        pcr.extra_percent = self.extra_percent
        
        if self.num_reactions:
            pcr.num_reactions = self.num_reactions
        if self.reaction_volume_uL:
            pcr.hold_ratios.volume = self.reaction_volume_uL, 'µL'

        pcr['water'].order = 1

        pcr['template DNA'].order = 2
        pcr['template DNA'].name = merge_names(self.template_tags)
        pcr['template DNA'].master_mix = all_equal(self.template_tags)

        if x := self.template_volume_uL:
            pcr['template DNA'].volume = x, 'µL'
        if x := self.template_stock:
            pcr['template DNA'].stock_conc = x

        # Setup the primers.  This is complicated because the primers might 
        # get split into their own mix, if the volumes that would be added 
        # to the PCR reaction are too small.

        primer_mix = None
        primer_keys = [
                'forward primer',
                'reverse primer',
        ]
        use_primer_mix = []

        for p, primers in zip(primer_keys, zip(*self.primer_pairs)):
            pcr[p].order = 3
            pcr[p].name = merge_names(x.tag for x in primers)
            pcr[p].hold_conc.stock_conc = min(x.stock_uM for x in primers), 'µM'
            pcr[p].master_mix = all_equal(x.tag for x in primers)

            primer_scale = pcr.scale if pcr[p].master_mix else 1
            primer_vol = primer_scale * pcr[p].volume

            if primer_vol < '0.5 µL':
                use_primer_mix.append(p)

        if use_primer_mix:
            pcr['primer mix'].order = 4
            pcr['primer mix'].stock_conc = '10x'
            pcr['primer mix'].volume = pcr.volume / 10
            pcr['primer mix'].master_mix = all(
                    pcr[p].master_mix
                    for p in use_primer_mix
            )

            primer_mix = stepwise.MasterMix()
            primer_mix.volume = pcr.volume

            for p in use_primer_mix:
                primer_mix[p].name = pcr[p].name
                primer_mix[p].stock_conc = pcr[p].stock_conc
                primer_mix[p].volume = pcr[p].volume
                primer_mix[p].hold_stock_conc.conc *= 10
                del pcr[p]

            primer_mix.hold_ratios.volume = '10 µL'

        return pcr, primer_mix

    def get_templates(self):
        return [x.template for x in self.amplicons]

    def get_template_tags(self):
        return [x.tag for x in self.templates]

    def get_primer_pairs(self):
        return [(x.fwd, x.rev) for x in self.amplicons]

    def get_product_seqs(self):
        return [x.seq for x in self.amplicons]

    def get_product_conc(self):
        return Quantity(self.product_conc_ng_uL, 'ng/µL')

    def get_product_volume(self):
        pcr, primers = self.reaction
        return pcr.volume

    def get_product_molecule(self):
        return 'dsDNA'

    @property
    def is_product_circular(self):
        return False

    def get_dependencies(self):
        return set(flatten(x.reagent_tags for x in self.amplicons))

if __name__ == '__main__':
    Pcr.main()




#!/usr/bin/env python3

import stepwise
import appcli
import autoprop

from stepwise import pl, ul
from stepwise_mol_bio import Assembly
from stepwise_mol_bio.digest import NebRestrictionEnzymeDatabase
from stepwise_mol_bio._assembly import ARGUMENT_DOC, OPTION_DOC
from freezerbox import MakerConfig, group_by_identity
from appcli import Key, DocoptConfig
from inform import plural

@autoprop.cache
class GoldenGate(Assembly):
    __doc__ = f"""\
Perform a Golden Gate assembly reaction.

Usage:
    golden_gate <assemblies>... [-c <conc>]... [-l <length>]... [options]

Arguments:
{ARGUMENT_DOC}

Options:
{OPTION_DOC}

    -e --enzymes <type_IIS>         [default: ${{','.join(app.enzymes)}}]
        The name(s) of the Type IIS restriction enzyme(s) to use for the 
        reaction.  To use more than one enzyme, enter comma-separated names.  
        Only NEB enzymes are supported.

Database:
    Golden gate assemblies can appear in the "Synthesis" column of a FreezerBox 
    database:
        
        golden-gate <assembly> [enzymes=<type IIS>] [volume=<µL>]

    <assembly>
        See the <assemblies>... command-line argument.

    volume=<µL>
        See --volume.  You must include a unit.

    enzymes=<type IIS>
        See --enzymes.
"""

    enzymes = appcli.param(
            Key(DocoptConfig, '--enzymes'),
            Key(MakerConfig, 'enzymes'),
            cast=lambda x: x.split(','),
            default=['BsaI-HFv2'],
    )

    group_by = {
            **Assembly.group_by,
            'enzymes': group_by_identity,
    }

    def get_reaction(self):
        rxn = stepwise.MasterMix()
        rxn.volume = '20 µL'

        rxn['T4 ligase buffer'].volume = '2.0 μL'
        rxn['T4 ligase buffer'].stock_conc = '10x'
        rxn['T4 ligase buffer'].master_mix = True
        rxn['T4 ligase buffer'].order = 2

        enz_uL = 0.5 if self.num_fragments <= 10 else 1.0

        rxn['T4 DNA ligase'].volume = enz_uL, 'µL'
        rxn['T4 DNA ligase'].stock_conc = '400 U/μL'
        rxn['T4 DNA ligase'].master_mix = True
        rxn['T4 DNA ligase'].order = 3

        enzyme_db = NebRestrictionEnzymeDatabase()

        for enzyme in self.enzymes:
            stock = enzyme_db[enzyme]['concentration'] / 1000
            rxn[enzyme].volume = enz_uL, 'µL'
            rxn[enzyme].stock_conc = stock, 'U/µL'
            rxn[enzyme].master_mix = True
            rxn[enzyme].order = 4

        return self._add_fragments_to_reaction(rxn)

    def del_reaction(self):
        pass

    def get_protocol(self):
        # Maybe this should be the getter function for the assembly param...

        p = stepwise.Protocol()
        rxn = self.reaction
        n = rxn.num_reactions
        n_frags = self.num_fragments

        f = 'https://tinyurl.com/yaa5mqz5'
        p += pl(
                f"Setup {plural(n):# Golden Gate assembl/y/ies}{p.add_footnotes(f)}:",
                rxn,
        )
        if n_frags <= 2:
            p += pl(
                    "Run the following thermocycler protocol:",
                    ul(
                        "37°C for 5 min",
                    ),
                    "Or, to maximize the number of transformants:",
                    ul(
                        "37°C for 60 min",
                        "60°C for 5 min",
                    ),
            )

        elif n_frags <= 4:
            p += pl(
                    "Run the following thermocycler protocol:",
                    ul(
                        "37°C for 60 min",
                        "60°C for 5 min",
                    ),
            )

        elif n_frags <= 10:
            p += pl(
                    "Run the following thermocycler protocol:",
                    ul(
                        "Repeat 30 times:",
                        ul(
                            "37°C for 1 min",
                            "16°C for 1 min",
                        ),
                        "60°C for 5 min",
                    ),
            )

        else:
            p += pl(
                    "Run the following thermocycler protocol:",
                    ul(
                        "Repeat 30 times:",
                        ul(
                            "37°C for 5 min",
                            "16°C for 5 min",
                        ),
                        "60°C for 5 min",
                    ),
            )

        return p

    def del_protocol(self):
        pass

if __name__ == '__main__':
    GoldenGate.main()



#!/usr/bin/env python3
# vim: tw=50

import stepwise
import autoprop
import appcli

from stepwise import pl, ul
from stepwise_mol_bio import Assembly
from stepwise_mol_bio._assembly import ARGUMENT_DOC, OPTION_DOC
from inform import plural

@autoprop.cache
class Ligate(Assembly):
    __doc__ = f"""\
Assemble restriction-digested DNA fragments using T4 DNA ligase.

Usage:
    ligate <assemblies>... [-c <conc>]... [-l <length>]... [options]

Arguments:
{ARGUMENT_DOC}

Options:
{OPTION_DOC}

    -k --kinase
        Add T4 polynucleotide kinase (PNK) to the reaction.  This is necessary 
        to ligate ends that are not already 5' phosphorylated (e.g. annealed 
        oligos, PCR products).

Database:
    Ligation reactions can appear in the "Synthesis" column of a FreezerBox 
    database:
        
        ligate <assembly> [volume=<µL>] [kinase=<bool>]

    <assembly>
        See the <assemblies>... command-line argument.

    volume=<µL>
        See --volume.  You must include a unit.

    kinase=<bool>
        See --kinase.  Specify "yes" or "no".
"""

    excess_insert = appcli.param(
            '--excess-insert',
            cast=float,
            default=3,
    )
    use_kinase = appcli.param(
            '--kinase',
            default=False,
    )

    def get_reaction(self):
        rxn = stepwise.MasterMix.from_text('''\
        Reagent               Stock        Volume  Master Mix
        ================  =========   ===========  ==========
        water                         to 20.00 μL         yes
        T4 ligase buffer        10x       2.00 μL         yes
        T4 DNA ligase      400 U/μL       1.00 μL         yes
        T4 PNK              10 U/μL       1.00 μL         yes
        ''')
        if not self.use_kinase:
            del rxn['T4 PNK']

        return self._add_fragments_to_reaction(rxn)

    def del_reaction(self):
        pass

    def get_protocol(self):
        p = stepwise.Protocol()
        rxn = self.reaction

        p += pl(
                f"Setup {plural(rxn.num_reactions):# ligation reaction/s}{p.add_footnotes('https://tinyurl.com/y7gxfv5m')}:",
                rxn,
        )
        p += pl(
                "Incubate at the following temperatures:",
                ul(
                    "25°C for 15 min",
                    "65°C for 10 min",
                ),
        )
        return p

    def del_protocol(self):
        pass

if __name__ == '__main__':
    Ligate.main()



#!/usr/bin/env python3

import stepwise
import autoprop

from stepwise import pl, ul
from stepwise_mol_bio import Assembly
from stepwise_mol_bio._assembly import ARGUMENT_DOC, OPTION_DOC
from inform import plural

@autoprop.cache
class Gibson(Assembly):
    __doc__ = f"""\
Perform a Gibson assembly reaction, using the NEB master mix (E2611).

Usage:
    gibson <assemblies>... [-c <conc>]... [-l <length>]... [options]

Arguments:
{ARGUMENT_DOC}

Options:
{OPTION_DOC}

Database:
    Gibson assemblies can appear in the "Synthesis" column of a FreezerBox 
    database:
        
        gibson <assembly> [volume=<µL>]

    <assembly>
        See the <assemblies>... command-line argument.

    volume=<µL>
        See --volume.  You must include a unit.
"""

    def get_reaction(self):
        rxn = stepwise.MasterMix()
        rxn.volume = '20 µL'

        rxn['Gibson master mix'].volume = rxn.volume / 2
        rxn['Gibson master mix'].stock_conc = "2x"
        rxn['Gibson master mix'].catalog_num = 'NEB E2611'
        rxn['Gibson master mix'].master_mix = True
        rxn['Gibson master mix'].order = 2

        return self._add_fragments_to_reaction(rxn)

    def del_reaction(self):
        pass

    def get_protocol(self):
        p = stepwise.Protocol()
        rxn = self.reaction
        n = rxn.num_reactions

        f = "https://tinyurl.com/ychbvkra"
        p += pl(
                f"Setup {plural(n):# Gibson assembl/y/ies}{p.add_footnotes(f)}:",
                rxn,
        )

        incubation_time = '15 min' if self.num_fragments <= 3 else '1h'
        p += f"Incubate at 50°C for {incubation_time}."
        return p

    def del_protocol(self):
        pass

if __name__ == '__main__':
    Gibson.main()

# vim: tw=50



#!/usr/bin/env python3
# vim: tw=50

import stepwise, appcli, autoprop
from inform import plural
from stepwise import pl, ul
from stepwise_mol_bio import Main

@autoprop
class Kld(Main):
    """\
Circularize a linear DNA molecule using T4 DNA ligase, e.g. to reform a plasmid 
after inverse PCR.

Usage:
    kld <dna> [-n <int>]

Arguments:
    <dna>
        The name of the DNA molecule to circularize.

Options:
    -n --num-reactions <int>        [default: ${app.num_reactions}]
        The number of reactions to set up.
"""
    __config__ = [
            appcli.DocoptConfig,
    ]

    dna = appcli.param('<dna>')
    num_reactions = appcli.param(
            '--num-reactions',
            cast=lambda x: int(eval(x)),
            default=1,
    )

    def __init__(self, dna):
        self.dna = dna

    def get_reaction(self):
        kld = stepwise.MasterMix.from_text('''\
        Reagent               Stock        Volume  Master Mix
        ================  =========   ===========  ==========
        water                         to 10.00 μL         yes
        T4 ligase buffer        10x       1.00 μL         yes
        T4 PNK              10 U/μL       0.25 μL         yes
        T4 DNA ligase      400 U/μL       0.25 μL         yes
        DpnI                20 U/μL       0.25 μL         yes
        DNA                50 ng/μL       1.50 μL
        ''')

        kld.num_reactions = self.num_reactions
        kld.extra_percent = 15
        kld['DNA'].name = self.dna

        return kld

    def get_protocol(self):
        p = stepwise.Protocol()
        p += pl(
                f"Setup {plural(self.num_reactions):# ligation reaction/s}:",
                self.reaction,
        )
        p += "Incubate at room temperature for 1h."
        return p

if __name__ == '__main__':
    Kld.main()



#!/usr/bin/env python3

from stepwise.testing import check_command, disable_capture
from param_helpers import *
from freezerbox.stepwise import Make

@parametrize_from_file(schema=app_expected_protocol)
def test_python_protocol(app, expected):
    assert match_protocol(app, expected)

@pytest.mark.slow
@parametrize_from_file(schema=cmd_stdout_stderr)
def test_cli_protocol(cmd, stdout, stderr):
    check_command(cmd, stdout=stdout)

@parametrize_from_file(schema=db_tags_expected_protocol)
def test_freezerbox_protocol(db, tags, expected, disable_capture):
    app = Make(db, tags)
    assert match_protocol(app, expected, disable_capture)

@parametrize_from_file(schema=db_expected)
def test_freezerbox_attrs(db, expected):
    for tag in expected:
        assert db[tag].dependencies == expected[tag]['dependencies']
        assert db[tag].conc == expected[tag]['conc']






#!/usr/bin/env python3

import stepwise, appcli, autoprop
from appcli import Key, Method, DocoptConfig
from stepwise import UsageError, StepwiseConfig, PresetConfig, pl, ul, dl
from stepwise_mol_bio import Main, ConfigError
from inform import plural

def parse_num_samples(name):
    try:
        return int(name)
    except ValueError:
        return len(name.strip(',').split(','))

def parse_sample_name(name):
    try:
        int(name)
        return None
    except ValueError:
        return name

@autoprop.cache
class Gel(Main):
    """\
Load, run, and stain gels.

Usage:
    gel <preset> <samples> [options]

<%! from stepwise_mol_bio import hanging_indent %>\
Arguments:
    <preset>
        What kind of gel to run.  The following presets are available:

        ${hanging_indent(app.preset_briefs, 8*' ')}

    <samples>
        The names of the samples to run, separated by commas.  This can also be 
        a number, which will be taken as the number of samples to run.

Options:
    -p --percent <number>
        The percentage of polyacrylamide/agarose in the gel being run.

    -a --additive <str>
        An extra component to include in the gel itself, e.g. 1x EtBr.

    -b --buffer <str>
        The buffer to run the gel in, e.g. TAE.

    -c --sample-conc <value>
        The concentration of the sample.  This will be used to scale how much 
        sample is mixed with loading buffer, with the goal of mixing the same 
        quantity of material specified in the preset.  In order to use this 
        option, the preset must specify a sample concentration.  The units of 
        that concentration will be used for this concentration.

    -v --sample-volume <µL>
        The volume of sample to mix with loading buffer, in µL.  This does not 
        scale the concentration, and may increase or decrease the amount of 
        sample loaded relative to what's specified in the preset.

    --mix-volume <µL>
        The volume of the sample/loading buffer mix to prepare for each sample.  
        For example, if you want to run two gels, but the preset only makes 
        enough mix for one, use this option to make more.

    --mix-extra <percent>
        How much extra sample/loading buffer mix to make.

    -M --no-mix
        Don't describe how to prepare the sample/loading buffer mix.

    --incubate-temp <°C>
        What temperature to incubate the sample/loading buffer at before 
        loading it onto the gel.  The incubation step will be skipped if 
        neither `--incubate-temp` nor `--incubate-time` are specified (either 
        on the command-line or via the preset).

    --incubate-time <min>
        How long to incubate the sample/loading buffer at the specified 
        temperature before loading it onto the gel.  The incubation step will 
        be skipped if neither `--incubate-temp` nor `--incubate-time` are 
        specified (either on the command-line or via the preset).

    --prerun-volts
        The voltage to pre-run the gel at.  By default, this will be the same 
        as the voltage the gel is run at.

    --prerun-time
        How long to prerun the gel, in minutes.

    -l --load-volume <µL>
        The volume of the sample/loading buffer mix to load onto the gel.

    --run-volts <V>
        The voltage to run the gel at.

    -r --run-time <min>
        How long to run the gel, in minutes.

    -s --stain <command>
        The name (and arguments) of a protocol describing how to stain the gel.  
        For example, this could be 'gelred' or 'coomassie -f'.
        
    -S --no-stain
        Don't describe how to stain/visualize the gel.

Configuration:
    Default values for this protocol can be specified in any of the following 
    stepwise configuration files:

        ${hanging_indent(app.config_paths, 8)}

    molbio.gel.presets:
        Named groups of default reaction parameters.  Typically each preset 
        corresponds to a particular kit or protocol.  See below for the various 
        settings that can be specified in each preset.

    molbio.gel.presets.<name>.title:
        How to briefly describe the gel in the protocol.  The default is 
        "electrophoresis".

    molbio.gel.presets.<name>.inherit:
        Copy all settings from another preset.  This can be used to make small 
        tweaks to a gel protocol, e.g. "SDS PAGE at lower-than-usual voltage".

    molbio.gel.presets.<name>.sample_mix:
        A table describing how to prepare the samples for the gel, in the 
        format understood by `stepwise.MasterMix.from_string()`.  The table 
        must contain one reagent named "sample".  This will be replaced with 
        the actual sample name specified on the command line, if possible.  If 
        no table is specified, the sample mix step will be left out of the 
        protocol.

    molbio.gel.presets.<name>.sample_conc:
        The default value for the `--sample-conc` option.

    molbio.gel.presets.<name>.sample_volume_uL:
        The default value for the `--sample-volume` option.

    molbio.gel.presets.<name>.ladder:
        The name of the ladder to use with this gel.

    molbio.gel.presets.<name>.ladder_volume_uL:
        How much ladder to load, in µL.

    molbio.gel.presets.<name>.mix_volume_uL:
        The default value for the `--mix-volume` option.
        
    molbio.gel.presets.<name>.mix_extra_percent:
        The default value for the `--mix-extra` option.

    molbio.gel.presets.<name>.incubate_temp_C:
        The default value for the `--incubate-temp` option.

    molbio.gel.presets.<name>.incubate_time_min:
        The default value for the `--incubate-time` option.

    molbio.gel.presets.<name>.gel_type:
        What kind of gel to use, e.g. "Bis-Tris/MES SDS PAGE", "TBE/urea PAGE", 
        "TAE/agarose", etc.  Don't include the gel percentage here; use the 
        `gel_percent` setting for that.

    molbio.gel.presets.<name>.gel_percent:
        The default value for the `--percent` option.

    molbio.gel.presets.<name>.gel_additive:
        The default value for the `--additive` option.

    molbio.gel.presets.<name>.gel_buffer:
        The default value for the `--buffer` option.
        
    molbio.gel.presets.<name>.load_volume_uL:
        The default value for the `--load-volume` option.

    molbio.gel.presets.<name>.prerun_volts:
        The default value for the `--prerun-volts` option.

    molbio.gel.presets.<name>.prerun_time_min:
        The default value for the `--prerun-time` option.

    molbio.gel.presets.<name>.run_volts:
        The default value for the `--run-volts` option.

    molbio.gel.presets.<name>.run_time_min:
        The default value for the `--run-time` option.

    molbio.gel.presets.<name>.stain:
        The default value for the `--stain` option.  If unspecified, there will 
        be no staining step by default.

    molbio.gel.presets.<name>.protocol_link:
        A hyperlink to an online description of the protocol, e.g. from the gel 
        manufacturer.  This link will be included as a footnote.
"""

    __config__ = [
            DocoptConfig,
            PresetConfig,
            StepwiseConfig.setup('molbio.gel'),
    ]
    preset_briefs = appcli.config_attr()
    config_paths = appcli.config_attr()

    presets = appcli.param(
            Key(StepwiseConfig, 'presets'),
            pick=list,
    )
    preset = appcli.param(
            Key(DocoptConfig, '<preset>'),
    )
    title = appcli.param(
            Key(PresetConfig, 'title'),
            default='electrophoresis',
    )
    num_samples = appcli.param(
            Key(DocoptConfig, '<samples>', cast=parse_num_samples),
            ignore=None,
            default=1,
    )
    sample_name = appcli.param(
            Key(DocoptConfig, '<samples>', cast=parse_sample_name),
            default=None,
    )
    sample_mix_str = appcli.param(
            Key(DocoptConfig, '--no-mix', cast=lambda x: None),
            Key(PresetConfig, 'sample_mix'),
            default=None,
    )
    sample_conc = appcli.param(
            Key(DocoptConfig, '--sample-conc'),
            Key(PresetConfig, 'sample_conc'),
            cast=float,
            default=None,
    )
    sample_volume_uL = appcli.param(
            Key(DocoptConfig, '--sample-volume'),
            Key(PresetConfig, 'sample_volume_uL'),
            cast=float,
            default=None,
    )
    ladder_name = appcli.param(
            Key(PresetConfig, 'ladder'),
            default=None,
    )
    ladder_volume_uL = appcli.param(
            Key(PresetConfig, 'ladder_volume_uL'),
    )
    mix_volume_uL = appcli.param(
            Key(DocoptConfig, '--mix-volume'),
            Key(PresetConfig, 'mix_volume_uL'),
            cast=float,
            default=None,
    )
    mix_extra_percent = appcli.param(
            Key(DocoptConfig, '--mix-extra'),
            Key(PresetConfig, 'mix_extra_percent'),
            cast=float,
            default=50,
    )
    incubate_temp_C = appcli.param(
            Key(DocoptConfig, '--incubate-temp'),
            Key(PresetConfig, 'incubate_temp_C'),
            cast=float,
    )
    incubate_time_min = appcli.param(
            Key(DocoptConfig, '--incubate-time'),
            Key(PresetConfig, 'incubate_time_min'),
            cast=int,
    )
    gel_type = appcli.param(
            Key(PresetConfig, 'gel_type'),
    )
    gel_percent = appcli.param(
            Key(DocoptConfig, '--percent'),
            Key(PresetConfig, 'gel_percent'),
    )
    gel_additive = appcli.param(
            Key(DocoptConfig, '--additive'),
            Key(PresetConfig, 'gel_additive'),
            default=None,
    )
    gel_buffer = appcli.param(
            Key(DocoptConfig, '--buffer'),
            Key(PresetConfig, 'gel_buffer'),
    )
    load_volume_uL = appcli.param(
            Key(DocoptConfig, '--load-volume'),
            Key(PresetConfig, 'load_volume_uL'),
            cast=float,
    )
    prerun_volts = appcli.param(
            Key(DocoptConfig, '--prerun-volts'),
            Key(PresetConfig, 'prerun_volts'),
            Method(lambda self: self.run_volts),
            cast=float,
    )
    prerun_time_min = appcli.param(
            Key(DocoptConfig, '--prerun-time'),
            Key(PresetConfig, 'prerun_time_min'),
            cast=int,
            default=None,
    )
    run_volts = appcli.param(
            Key(DocoptConfig, '--run-volts'),
            Key(PresetConfig, 'run_volts'),
            cast=float,
    )
    run_time_min = appcli.param(
            Key(DocoptConfig, '--run-time'),
            Key(PresetConfig, 'run_time_min'),
            cast=int,
    )
    stain = appcli.param(
            Key(DocoptConfig, '--stain'),
            Key(DocoptConfig, '--no-stain', cast=lambda x: None),
            Key(PresetConfig, 'stain'),
            default=None,
    )
    protocol_link = appcli.param(
            Key(PresetConfig, 'protocol_link'),
            default=None,
    )

    def __init__(self, preset, num_samples=None):
        self.preset = preset
        self.num_samples = num_samples

    def get_protocol(self):
        p = stepwise.Protocol()

        if self.sample_mix:
            p += self.prep_step

        p += self.run_step
            
        if self.stain:
            p += stepwise.load(self.stain)

        return p

    def del_protocol(self):
        pass

    def get_prep_step(self):

        def both_or_neither(key1, key2):
            has_key1 = has_key2 = True

            try: value1 = getattr(self, key1)
            except AttributeError: has_key1 = False

            try: value2 = getattr(self, key2)
            except AttributeError: has_key2 = False

            if has_key1 and not has_key2:
                raise ConfigError(f"specified {key1!r} but not {key2!r}")
            if has_key2 and not has_key1:
                raise ConfigError(f"specified {key2!r} but not {key1!r}")

            if has_key1 and has_key2:
                return value1, value2
            else:
                return False

        s = pl(
                f"Prepare {plural(self.num_samples):# sample/s} for {self.title}:",
                self.sample_mix,
        )
        if x := both_or_neither('incubate_temp_C', 'incubate_time_min'):
            temp_C, time_min = x
            s += ul(
                    f"Incubate at {temp_C:g}°C for {time_min:g} min."
            )

        return s

    def del_prep_step(self):
        pass

    def get_run_step(self):
        p = stepwise.Protocol()
        additive = f" with {x}" if (x := self.gel_additive) else ""
        percent = x.replace('-', '–') if isinstance(x := self.gel_percent, str) else x
        p += pl(f"Run a gel{p.add_footnotes(self.protocol_link)}:", dl(
            ("gel", f"{percent}% {self.gel_type}{additive}"),
            ("buffer", f"{self.gel_buffer}"),
            ("ladder", self.ladder_name and f"{self.ladder_volume_uL:g} µL {self.ladder_name}"),
            ("samples", self.load_volume_uL and f"{self.load_volume_uL:.3g} µL/lane"),
            ("prerun", self.prerun_time_min and f"{self.prerun_volts:g}V for {self.prerun_time_min:g} min"),
            ("run", f"{self.run_volts:g}V for {self.run_time_min:g} min"),
        ))
        return p

    def del_run_step(self):
        pass

    def get_sample_mix(self):
        if not self.sample_mix_str:
            return None

        mix = stepwise.MasterMix.from_text(self.sample_mix_str)
        mix.num_reactions = self.num_samples
        mix.extra_percent = self.mix_extra_percent
        mix['sample'].name = self.sample_name

        if x := self.sample_conc:
            stock_conc = mix['sample'].stock_conc
            if stock_conc is None:
                raise ConfigError(f"can't change sample stock concentration, no initial concentration specified.")
            mix['sample'].hold_conc.stock_conc = x, stock_conc.unit

        if x := self.sample_volume_uL:
            mix['sample'].volume = x, 'µL'

        if x := self.mix_volume_uL:
            mix.hold_ratios.volume = x, 'µL'

        if mix.solvent:
            mix.fix_volumes('sample')

        return mix

    def del_sample_mix(self):
        pass

if __name__ == '__main__':
    Gel.main()



#!/usr/bin/env python3

import pytest
import parametrize_from_file

from stepwise.testing import disable_capture
from stepwise_mol_bio.pcr import *
from freezerbox.stepwise import Make
from pytest import approx
from param_helpers import *

@parametrize_from_file(
        schema=Schema({
            str: And(str, str.strip),
            'is_circular': eval,
            **error_or({
                'expected': And(str, str.strip),
            }),
        }),
)
def test_find_amplicon(template, primer_1, primer_2, is_circular, expected, error):
    with error:
        amplicon = find_amplicon(template, primer_1, primer_2, is_circular)
        assert amplicon == expected.upper()

@parametrize_from_file(
        schema=Schema({
            'given': str,
            **error_or({
                'expected': {str: str},
            }),
        })
)
def test_parse_amplicon(given, expected, error):
    with error:
        amplicon = parse_amplicon(given)
        assert amplicon.template.tag == expected['template']
        assert amplicon.fwd.tag == expected['fwd']
        assert amplicon.rev.tag == expected['rev']

@parametrize_from_file(
        schema=Schema({
            'given': str,
            **error_or({
                'expected': eval_with(),
            }),
        })
)
def test_parse_primers(given, expected, error):
    with error:
        assert parse_primers(given) == expected

@parametrize_from_file(
        schema=Schema({
            'app': exec_app,
            'expected': {str: str},
        }),
)
def test_reaction(app, expected):
    pcr, primers = app.reaction

    assert str(pcr) == expected['pcr']

    if 'primers' in expected:
        assert str(primers) == expected['primers']
    else:
        assert primers is None

@parametrize_from_file(schema=app_expected_error)
def test_product_seqs(app, expected, error):
    with error:
        assert app.product_seqs == [x.upper() for x in expected]

@parametrize_from_file(schema=app_expected)
def test_anneal_temp_C(app, expected):
    assert app.anneal_temp_C == approx(expected)

@parametrize_from_file(schema=app_expected)
def test_extend_time_s(app, expected):
    assert app.extend_time_s == approx(expected)

@parametrize_from_file(
        key='test_freezerbox_make',
        schema=db_tags_expected_protocol,
)
def test_protocol_products(db, tags, expected):
    if len(tags) != 1:
        pytest.skip()

    app = Pcr.from_product(tags[0])
    app.db = db

    assert match_protocol(app, expected)

@parametrize_from_file(schema=db_tags_expected_protocol)
def test_freezerbox_make(db, tags, expected, disable_capture):
    app = Make(db, tags)
    assert match_protocol(app, expected, disable_capture)

@parametrize_from_file(schema=db_expected)
def test_freezerbox_attrs(db, expected):
    for tag in expected:
        assert db[tag].seq == expected[tag]['seq'].upper()
        assert db[tag].dependencies == expected[tag]['dependencies']
        assert db[tag].conc_ng_uL == expected[tag]['conc_ng_uL']
        assert db[tag].volume_uL == expected[tag]['volume_uL']
        assert db[tag].molecule == 'DNA'
        assert db[tag].is_double_stranded == True
        assert db[tag].is_circular == False




#!/usr/bin/env python3

from stepwise_mol_bio._assembly import *
from param_helpers import *

@parametrize_from_file(
        schema=Schema({
            'args': [str],
            Optional('parse_value', default='lambda x: x'): eval,
            'expected': {str: eval},
        }),
)
def test_parse_fragment_attrs(args, parse_value, expected):
    actual = parse_fragment_attrs(args, parse_value)
    for k, v in expected.items():
        assert actual[k] == v

@parametrize_from_file(
        schema=Schema({
            'args': {str: eval_with()},
            **error_or({
                'expected': eval_swmb,
            })
        }),
)
def test_parse_assemblies_from_docopt(args, expected, error):
    with error:
        assert parse_assemblies_from_docopt(args) == expected

@parametrize_from_file(
        schema=Schema({
            'assemblies': [[eval_swmb]],
            Optional('kwargs', default={}): {str: eval_swmb},
            'expected': str,
            Optional('warning', default=''): str,
        }),
)
def test_add_fragments_to_reaction(assemblies, kwargs, expected, warning, capsys):
    rxn = stepwise.MasterMix()
    rxn.volume = 5, 'µL'

    add_fragments_to_reaction(rxn, assemblies, **kwargs)
    cap = capsys.readouterr()

    assert str(rxn) == expected
    assert warning in cap.err



#!/usr/bin/env python3

import parametrize_from_file
from param_helpers import *

@parametrize_from_file(schema=app_expected_protocol)
def test_protocol(app, expected):
    assert match_protocol(app, expected)




#!/usr/bin/env python3

import sys
import appcli
import autoprop
import tidyexc

from freezerbox import MakerConfig, iter_combo_makers
from appcli import Method, DocoptConfig
from appdirs import AppDirs
from inform import format_range, error
from more_itertools import all_equal
from functools import partial
from pathlib import Path

app_dirs = AppDirs("stepwise_mol_bio")

class Main(appcli.App):
    usage_io = sys.stderr
    group_by = {}
    merge_by = {}

    @classmethod
    def main(cls):
        app = cls.from_params()
        app.load(DocoptConfig)
        app.load(MakerConfig)
        
        try:
            app.protocol.print()
        except StepwiseMolBioError as err:
            error(err)

    @classmethod
    def make(cls, db, products, *, group_by=None, merge_by=None):
        if group_by is None:
            group_by = cls.group_by

        if merge_by is None:
            merge_by = cls.merge_by

        yield from iter_combo_makers(
                partial(cls._combo_maker_factory, db),
                map(cls._solo_maker_factory, products),
                group_by=group_by,
                merge_by=merge_by,
        )

    def refresh(self):
        autoprop.clear_cache(self)

    @classmethod
    def _solo_maker_factory(cls, product):
        app = cls.from_params()
        app.db = product.db
        app.products = [product]
        app.load(MakerConfig)
        return app

    @classmethod
    def _combo_maker_factory(cls, db):
        app = cls.from_params()
        app.db = db
        return app


class Cleanup(Main):

    product_tags = appcli.param(
            Method(lambda self: [x.tag for x in self.products]),
            default_factory=list,
    )

    def __bareinit__(self):
        super().__bareinit__()
        self.show_product_tags = False

    @classmethod
    def make(cls, db, products):
        makers = list(super().make(db, products))
        show_product_tags = (len(makers) != 1)

        for maker in makers:
            maker.show_product_tags = show_product_tags
            yield maker


@autoprop
class Argument(appcli.App):

    def __init__(self, tag, **kwargs):
        self.tag = tag
        self._set_known_attrs(kwargs)

    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.tag!r})'

    def bind(self, app, force=False):
        if not hasattr(self, 'app') or force:
            self.app = app
            self.on_bind(app)

    def on_bind(self, app):
        pass

    def get_db(self):
        return self.app.db

    def _set_known_attrs(self, attrs):
        for attr, value in attrs.items():
            if attr not in self.__class__.__dict__:
                raise AttributeError(f"unknown attribute {attr!r}")
            setattr(self, attr, value)


class ShareConfigs:

    def on_bind(self, app):
        appcli.share_configs(app, self)



class StepwiseMolBioError(tidyexc.Error):
    pass

class ConfigError(StepwiseMolBioError):
    # For values that don't make sense, e.g. non-existent enzymes, etc.
    pass

class UsageError(StepwiseMolBioError):
    # For if the program isn't being used correctly, e.g. missing information.
    pass

def bind_arguments(app, reagents, iter=iter):
    for reagent in iter(reagents):
        reagent.bind(app)
    return reagents

def try_except(expr, exc, failure, success=None):
    try:
        x = expr()
    except exc:
        return failure()
    else:
        return success() if success else x

def hanging_indent(text, prefix):
    from textwrap import indent
    if not isinstance(text, str):
        text = '\n'.join(map(str, text))
    if isinstance(prefix, int):
        prefix = ' ' * prefix
    return indent(text, prefix)[len(prefix):]

def merge_dicts(dicts):
    result = {}
    for dict in reversed(list(dicts)):
        result.update(dict)
    return result

def comma_list(x):
    return [x.strip() for x in x.split(',')]

def comma_set(x):
    return {x.strip() for x in x.split(',')}

def int_or_expr(x):
    return type_or_expr(int, x)

def float_or_expr(x):
    return type_or_expr((float, int), x)

def type_or_expr(type, x):
    if isinstance(x, type):
        return x
    else:
        return type(eval(x))

def require_reagent(rxn, reagent):
    if reagent not in rxn:
        raise UsageError(f"reagent table missing {reagent!r}")

def merge_names(names):
    names = list(names)
    if all_equal(names):
        return names[0]
    else:
        return ','.join(names)

def match_len(x, n):
    # Something more generic than this might fit well in `more_itertools`.
    if isinstance(x, list):
        if len(x) != n:
            raise ValueError(f"expected {n} item(s), got {len(x)}")
        return x
    else:
        return n * [x]


def format_sec(x):
    if x < 60:
        return f'{x}s'

    min = x // 60
    sec = x % 60

    return f'{min}m{f"{sec:02}" if sec else ""}'

def format_min(x):
    if x < 60:
        return f'{x}m'

    hr = x // 60
    min = x % 60

    return f'{hr}h{f"{sec:02}" if min else ""}'




#!/usr/bin/env python3

import stepwise, appcli, autoprop
from stepwise import Quantity
from stepwise_mol_bio import Cleanup
from freezerbox import MakerConfig, group_by_identity
from appcli import DocoptConfig, Key, Method

@autoprop
class Aliquot(Cleanup):
    """\
Make aliquots

Usage:
    aliquot <volume> [<conc>]

Arguments:
    <volume>
        The volume of each individual aliquot.  No unit is implied, so you 
        must specify one.

    <conc>
        The concentration of the aliquots, if this is not made clear in 
        previous steps.  No unit is implied, so you must specify one.
"""
    __config__ = [
            DocoptConfig,
            MakerConfig,
    ]
    volume = appcli.param(
            Key(DocoptConfig, '<volume>'),
            Key(MakerConfig, 'volume'),
    )
    conc = appcli.param(
            Key(DocoptConfig, '<conc>'),
            Key(MakerConfig, 'conc'),
            default=None,
    )

    group_by = {
        'volume': group_by_identity,
        'conc': group_by_identity,
    }

    def __init__(self, volume, conc=None, product_tags=None):
        self.volume = volume
        if conc: self.conc = conc
        if product_tags: self.product_tags = product_tags

    def get_protocol(self):
        Q = Quantity.from_string

        if self.conc:
            aliquot_info = f'{Q(self.volume)}, {Q(self.conc)}'
        else:
            aliquot_info = f'{Q(self.volume)}'

        if self.product_tags and self.show_product_tags:
            product_tags = f" of: {', '.join(self.product_tags)}"
        else:
            product_tags = "."

        return stepwise.Protocol(
                steps=[f"Make {aliquot_info} aliquots{product_tags}"],
        )

    def get_product_conc(self):
        return Quantity.from_string(self.conc)

if __name__ == '__main__':
    Aliquot.main()



#!/usr/bin/env python3

import stepwise
import appcli
import autoprop
import re
import json
import requests

from stepwise_mol_bio import (
        Main, Argument, ShareConfigs, UsageError, ConfigError,
        bind_arguments, app_dirs, comma_list, match_len, int_or_expr,
)
from stepwise import StepwiseConfig, pl, ul
from freezerbox import (
        ReagentConfig, MakerConfig,
        parse_mass_ug, parse_volume_uL, parse_size_bp, parse_time,
        group_by_identity, join_lists,
)
from appcli import Key, Method, DocoptConfig
from inform import Error, plural, did_you_mean
from functools import partial
from pathlib import Path

def parse_templates_from_csv(string):
    return [
            RestrictionDigest.Template(x)
            for x in comma_list(string)
    ]

def parse_template_from_freezerbox(string):
    return [RestrictionDigest.Template(string)]

def pick_compatible_buffer(enzymes):
    if len(enzymes) == 1:
        return enzymes[0]['recommBuffer']

    # Don't consider `buf5`.  This is the code for buffers that are unique to a 
    # specific enzyme, so even if two enzymes both want `buf5`, it's not the 
    # same buffer.

    buffer_names = {
            '1': "NEBuffer r1.1",
            '2': "NEBuffer r2.1",
            '3': "NEBuffer r3.1",
            '4': "rCutSmart Buffer",
    }
    buffer_scores = {
            k: (
                sum(not x[f'star{k}'] for x in enzymes),  # Star activity?
                sum(x[f'buf{k}'] for x in enzymes),       # Cutting activity?
                k == '4',                                 # Prefer CutSmart
            )
            for k in buffer_names
    }
    best_buffer = max(
            buffer_scores,
            key=lambda k: buffer_scores[k],
    )

    return buffer_names[best_buffer]

def calc_digest_products(seq, enzymes, *, is_circular):
    from more_itertools import pairwise, flatten
    from Bio.Restriction import RestrictionBatch
    from Bio.Seq import Seq

    if not enzymes:
        raise UsageError("no enzymes specified", enzymes=enzymes)

    enzymes = [
            re.sub('-HF(v2)?$', '', x)
            for x in enzymes
    ]

    try:
        batch = RestrictionBatch(enzymes)
    except ValueError:
        raise ConfigError(
                lambda e: f"unknown enzyme(s): {','.join(map(repr, e.enzymes))}",
                enzymes=enzymes,
        ) from None

    sites = [x-1 for x in flatten(batch.search(Seq(seq)).values())]

    if not sites:
        raise ConfigError(
                lambda e: f"{','.join(map(repr, e.enzymes))} {plural(enzymes):/does/do} not cut template.",
                enzymes=enzymes,
                seq=seq,
        )

    sites += [] if is_circular else [0, len(seq)]
    sites = sorted(sites)

    seqs = []
    for i,j in pairwise(sorted(sites)):
        seqs.append(seq[i:j])

    if is_circular:
        wrap_around = seq[sites[-1]:] + seq[:sites[0]]
        seqs.append(wrap_around)

    return seqs

def calc_digest_product(seq, enzymes, *, is_circular, target_size=None):
    target_size = target_size or len(seq)
    product_seqs = calc_digest_products(seq, enzymes, is_circular=is_circular)
    return min(product_seqs, key=lambda x: abs(target_size - len(x)))

@autoprop.cache
class RestrictionDigest(Main):
    """\
Perform restriction digests using the protocol recommended by NEB.

Usage:
    digest <templates> <enzymes> [-d <ng>] [-D <ng/µL>] [-v <µL>] [-n <rxns>]
        [-g]
    digest <product> [-e <enzymes>] [options]

Arguments:
    <templates>
        The DNA to digest.  Use commas to specify multiple templates.  The 
        number of reactions will equal the number of templates.

    <enzymes>
        The restriction enzymes to use.  Only NEB enzymes are currently 
        supported.  If you are using an "HF" enzyme, specify that explicitly.  
        For example, "HindIII" and "HindIII-HF" have different protocols.  
        Enzyme names are case-insensitive, and multiple enzymes can be 
        specified using commas.

    <product>
        A product in the FreezerBox database that was synthesized by 
        restriction digest.  If this form of the command is given, the protocol 
        will take all default values (including the template and enzymes) from 
        the reaction used to synthesize the given product.

Options:
    -d --dna <µg>               [default: ${app.dna_ug}]
        The amount of DNA to digest, in µg.

    -D --dna-stock <ng/µL>
        The stock concentration of the DNA template, in ng/µL.

    -v --target-volume <µL>     [default: ${app.target_volume_uL}]
        The ideal volume for the digestion reaction.  Note that the actual 
        reaction volume may be increased to ensure that the volume of enzyme 
        (which is determined by the amount of DNA to digest, see --dna) is less 
        than 10% of the total reaction volume, as recommended by NEB.

    -t --time <min>
        Incubate the digestion reaction for a non-standard amount of time.  You 
        may optionally specify a unit.  If you don't, minutes are assumed.

    -n --num-reactions <int>
        The number of reactions to setup.  By default, this is inferred from 
        the number of templates.

    -g --genomic
        Indicate that genomic DNA is being digested.  This will double the 
        amount of enzyme used, as recommended by NEB.

    -e --enzymes <list>
        The same as <enzymes>, but for when a product is specified.  
    """
    __config__ = [
            DocoptConfig,
            MakerConfig,
            StepwiseConfig.setup('molbio.digest'),
    ]

    class Template(ShareConfigs, Argument):
        __config__ = [ReagentConfig]

        seq = appcli.param(
                Key(ReagentConfig, 'seq'),
        )
        stock_ng_uL = appcli.param(
                Key(DocoptConfig, '--dna-stock'),
                Key(ReagentConfig, 'conc_ng_uL'),
                Key(StepwiseConfig, 'dna_stock_ng_uL'),
                cast=float,
        )
        is_circular = appcli.param(
                Key(ReagentConfig, 'is_circular'),
                default=True,
        )
        is_genomic = appcli.param(
                Key(DocoptConfig, '--genomic'),
                default=False,
        )
        target_size_bp = appcli.param(
                Key(MakerConfig, 'size', cast=parse_size_bp),
                default=None,
        )

    templates = appcli.param(
            Key(DocoptConfig, '<templates>', cast=parse_templates_from_csv),
            Key(MakerConfig, 'template', cast=parse_template_from_freezerbox),
            get=bind_arguments,
    )
    enzyme_names = appcli.param(
            Key(DocoptConfig, '<enzymes>', cast=comma_list),
            Key(DocoptConfig, '--enzymes', cast=comma_list),
            Key(MakerConfig, 'enzymes', cast=comma_list),
    )
    product_tag = appcli.param(
            Key(DocoptConfig, '<product>'),
    )
    products = appcli.param(
            Method(lambda self: [
                self.db[self.product_tag].make_intermediate(0)
            ]),
    )
    num_reactions = appcli.param(
            Key(DocoptConfig, '--num-reactions', cast=int_or_expr),
            Method(lambda self: len(self.templates)),
    )
    dna_ug = appcli.param(
            Key(DocoptConfig, '--dna', cast=partial(parse_mass_ug, default_unit='µg')),
            Key(MakerConfig, 'mass', cast=parse_mass_ug),
            Key(StepwiseConfig),
            cast=float,
            default=1,
    )
    target_volume_uL = appcli.param(
            Key(DocoptConfig, '--target-volume', cast=partial(parse_volume_uL, default_unit='µL')),
            Key(MakerConfig, 'volume', cast=parse_volume_uL),
            Key(StepwiseConfig),
            default=10,
    )
    target_size_bp = appcli.param(
            Key(MakerConfig, 'size', cast=parse_size_bp),
            default=None,
    )
    time = appcli.param(
            Key(DocoptConfig, '--time', cast=partial(parse_time, default_unit='min')),
            Key(MakerConfig, 'time', cast=parse_time),
            default=None,
    )

    group_by = {
        'enzyme_names': group_by_identity,
        'dna_ug': group_by_identity,
        'target_volume_uL': group_by_identity,
        'time': group_by_identity,
    }
    merge_by = {
        'templates': join_lists,
    }

    @classmethod
    def from_tags(cls, template_tags, enzyme_names, db=None):
        templates = [cls.Template(x) for x in template_tags]
        return cls(templates, enzyme_names, db)

    @classmethod
    def from_product(cls, product_tag):
        self = cls.from_params()
        self.product_tag = product_tag
        self.load(MakerConfig)
        return self


    def __bareinit__(self):
        self._enzyme_db = None

    def __init__(self, templates, enzyme_names, db=None):
        self.templates = templates
        self.enzyme_names = enzyme_names
        self.enzyme_db = db

    def get_enzymes(self):
        return [self.enzyme_db[x] for x in self.enzyme_names]

    def get_enzyme_db(self):
        if self._enzyme_db is None:
            self._enzyme_db = NebRestrictionEnzymeDatabase()
        return self._enzyme_db

    def set_enzyme_db(self, db):
        self._enzyme_db = db

    def get_reaction(self):
        # Define a prototypical restriction digest reaction.  Stock 
        # concentrations for BSA, SAM, and ATP come from the given catalog 
        # numbers.

        rxn = stepwise.MasterMix.from_text("""\
        Reagent   Catalog      Stock    Volume  MM?
        ========  =======  =========  ========  ===
        water                         to 50 µL  yes
        DNA                200 ng/µL      5 µL  yes
        buffer                   10x      5 µL  yes
        bsa         B9200   20 mg/mL      0 µL  yes
        sam         B9003      32 mM      0 µL  yes
        atp         P0756      10 mM      0 µL  yes
        """)

        # Plug in the parameters the user requested.

        rxn.num_reactions = self.num_reactions

        rxn['DNA'].name = ','.join(x.tag for x in self.templates)
        rxn['DNA'].hold_conc.stock_conc = min(
                x.stock_ng_uL for x in self.templates), 'ng/µL'

        if len(self.templates) > 1:
            rxn['DNA'].order = -1
            rxn['DNA'].master_mix = False
        
        for enz in self.enzymes:
            key = enz['name']
            stock = enz['concentration'] / 1000
            is_genomic = any(x.is_genomic for x in self.templates)

            # The prototype reaction has 1 µg of DNA.  NEB recommends 10 U/µg 
            # (20 U/µg for genomic DNA), so set the initial enzyme volume 
            # according to that.  This will be adjusted later on.

            rxn[key].stock_conc = stock, 'U/µL'
            rxn[key].volume = (20 if is_genomic else 10) / stock, 'µL'
            rxn[key].master_mix = True

        rxn['buffer'].name = pick_compatible_buffer(self.enzymes)

        # Supplements

        known_supplements = []

        def add_supplement(key, name, unit, scale=1):
            conc = max(x['supplement'][key] for x in self.enzymes)
            known_supplements.append(key)

            if not conc:
                del rxn[key]
            else:
                rxn[key].hold_stock_conc.conc = conc * scale, unit
                rxn[key].name = name

        add_supplement('bsa', 'rAlbumin', 'mg/mL', 1e-3)
        add_supplement('sam', 'SAM', 'mM', 1e-3)
        add_supplement('atp', 'ATP', 'mM')

        # Make sure there aren't any supplements we should add that we don't 
        # know about.
        
        for enzyme in self.enzymes:
            for supp, conc in enzyme['supplement'].items():
                if conc > 0 and supp not in known_supplements:
                    err = ConfigError(
                            enzyme=enzyme,
                            supp=supp,
                            conc=conc,
                    )
                    err.brief = "{enzyme[name]!r} requires an unknown supplement: {supp!r}"
                    err.hints += "the restriction digest protocol needs updated"
                    err.hints += "please submit a bug report"
                    raise err
        
        # Update the reaction volume.  This takes some care, because the 
        # reaction volume depends on the enzyme volume, which in turn depends 
        # on the DNA quantity.

        k = self.dna_ug / 1  # The prototype reaction has 1 µg DNA.
        dna_vol = k * rxn['DNA'].volume
        enz_vols = {
                enz['name']: k * rxn[enz['name']].volume
                for enz in self.enzymes
        }
        enz_vol = sum(enz_vols.values())

        rxn.hold_ratios.volume = max(
                stepwise.Quantity(self.target_volume_uL, 'µL'),
                10 * enz_vol,

                # This is a bit of a hack.  The goal is to keep the water 
                # volume non-negative, but it won't necessarily work if there 
                # are supplements.
                10/9 * (dna_vol + enz_vol),
        )

        rxn['DNA'].volume = dna_vol
        for enz in self.enzymes:
            key = enz['name']
            rxn[key].volume = enz_vols[key]

        return rxn

    def del_reaction(self):
        pass

    def get_protocol(self):
        from itertools import groupby
        from operator import itemgetter

        protocol = stepwise.Protocol()
        rxn = self.reaction
        rxn_type = (
                self.enzymes[0]['name']
                if len(self.enzymes) == 1 else
                'restriction'
        )

        def incubate(temp_getter, time_getter, time_formatter=lambda x: f'{x} min'):
            incubate_params = [
                    (k, max(time_getter(x) for x in group))
                    for k, group in groupby(self.enzymes, temp_getter)
            ]
            return [
                    f"{temp}°C for {time_formatter(time)}"
                    for temp, time in sorted(incubate_params)
            ]

        if self.time:
            digest_steps = incubate(
                    itemgetter('incubateTemp'),
                    lambda x: self.time,
                    lambda x: x,
            )
        else:
            digest_steps = incubate(
                    itemgetter('incubateTemp'),
                    lambda x: 15 if x['timeSaver'] else 60,
                    lambda x: '5–15 min' if x == 15 else '1 hour',
            )

        inactivate_steps = incubate(
                itemgetter('heatInactivationTemp'),
                itemgetter('heatInactivationTime'),
        )

        protocol += pl(
            f"Setup {plural(rxn.num_reactions):# {rxn_type} digestion/s} [1,2]:",
            rxn,
        )
        protocol += pl(
            f"Incubate at the following temperatures [3]:",
            ul(*digest_steps, *inactivate_steps),
        )

        urls = [x['url'] for x in self.enzymes if x.get('url')]
        protocol.footnotes[1] = pl(*urls, br='\n')
        protocol.footnotes[2] = """\
NEB recommends 5–10 units of enzyme per µg DNA 
(10–20 units for genomic DNA).  Enzyme volume 
should not exceed 10% of the total reaction 
volume to prevent star activity due to excess 
glycerol.
"""
        protocol.footnotes[3] = """\
The heat inactivation step is not necessary if 
the DNA will be purified before use.
"""
        return protocol

    def del_protocol(self):
        pass

    def get_dependencies(self):
        return {x.tag for x in self.templates}

    def get_product_seqs(self):
        for template in self.templates:
            with ConfigError.add_info("tag: {tag}", tag=template.tag):
                yield calc_digest_product(
                        seq=template.seq,
                        enzymes=self.enzyme_names,
                        target_size=template.target_size_bp,
                        is_circular=template.is_circular,
                )

    def get_product_conc(self):
        return self.reaction['DNA'].conc

    def get_product_volume(self):
        return self.reaction.volume


class NebRestrictionEnzymeDatabase:

    def __init__(self, cache_path=None):
        self.cache_path = cache_path or Path(app_dirs.user_cache_dir) / 'neb' / 'restriction_enzymes.json'
        self.load_cache()

    def __getitem__(self, name):
        try:
            return self.enzyme_params[name.lower()]

        except KeyError:
            try:
                self.download_cache()
                self.load_cache()
                return self.enzyme_params[name.lower()]

            except (KeyError, ConfigError) as err1:
                err2 = ConfigError(
                        enzyme=name,
                        known_enzymes=self.enzyme_names,
                        fresh_download=not isinstance(err1, ConfigError),
                )
                err2.brief = "no such enzyme {enzyme!r}"
                err2.info += lambda e: (
                        f"successfully downloaded the most recent restriction enzyme data from NEB (in case {e.enzyme!r} is a new enzyme)"
                        if e.fresh_download else
                        f"failed to download the most recent restriction enzyme data from NEB (in case {e.enzyme!r} is a new enzyme)"
                )
                err2.hints += lambda e: f"did you mean: {did_you_mean(e.enzyme, e.known_enzymes)!r}"

                raise err2 from err1

    def load_cache(self):
        if not self.cache_path.exists():
            self.download_cache()

        with self.cache_path.open() as f:
            data = json.load(f)

        self.enzyme_names = list(data.keys())
        self.enzyme_params = {k.lower(): v for k, v in data.items()}

    def download_cache(self):
        url = 'http://nebcloner.neb.com/data/reprop.json'

        try:
            data = requests.get(url).json()

        except requests.exceptions.ConnectionError as err1:
            err2 = ConfigError(url=url)
            err2.brief = "failed to download restriction enzyme data from NEB"
            err2.info += "URL: {url}"
            err2.hints += "make sure the internet is connected and the above URL is reachable."
            raise err2 from err1

        else:
            self.cache_path.parent.mkdir(parents=True, exist_ok=True)

            with self.cache_path.open('w') as f:
                json.dump(data, f)

if __name__ == '__main__':
    RestrictionDigest.main()



#!/usr/bin/env python3

"""
Protocols relating to molecular biology, e.g. PCR.
"""

__version__ = '1.23.0'

from ._utils import *
from ._assembly import Assembly
from .aliquot import Aliquot
from .anneal import Anneal
from .digest import RestrictionDigest
from .direct_dilution import DirectDilution
from .ethanol_precipitation import EthanolPrecipitation
from .gels.gel import Gel
from .gels.laser_scanner import LaserScanner
from .gels.stain import Stain
from .gels.uv_transilluminator import UvTransilluminator
from .gibson import Gibson
from .golden_gate import GoldenGate
from .ivt import Ivt
from .ivtt import InVitroTranslation
from .kld import Kld
from .ligate import Ligate
from .lyophilize import Lyophilize
from .miniprep import Miniprep
from .pcr import Pcr
from .serial_dilution import SerialDilution
from .spin_cleanup import SpinCleanup

# Avoid circular imports
from .invpcr import InversePcr
from .qpcr import Qpcr
from .page_purify import PagePurify

import stepwise
from pathlib import Path

class Plugin:
    protocol_dir = Path(__file__).parent
    config_path = protocol_dir / 'conf.toml'
    priority = stepwise.Builtins.priority + 10

del stepwise
del Path



#!/usr/bin/env python3

import autoprop
from stepwise import Protocol, Quantity
from stepwise_mol_bio import Cleanup
from appcli import DocoptConfig

@autoprop
class Miniprep(Cleanup):
    """
Purify plasmid by miniprep.

Usage:
    miniprep

Database:
    The miniprep protocol can be used in the "Cleanup" column of a FreezerBox 
    database:

        miniprep

    Miniprepped plasmids will be assumed to have a concentration of 200 ng/µL, 
    unless otherwise specified.
"""
    __config__ = [
            DocoptConfig,
    ]

    def get_protocol(self):
        p = Protocol()
        p += "Miniprep."
        return p

    def get_product_conc(self):
        return Quantity(200, 'ng/µL')

if __name__ == '__main__':
    Miniprep.main()



#!/usr/bin/env python3

import stepwise, appcli, autoprop
from inform import indent, plural
from appcli import DocoptConfig
from stepwise import StepwiseConfig, PresetConfig, UsageError, pl, pre
from stepwise_mol_bio import Main, ConfigError

@autoprop
class LaserScanner(Main):
    """\
Image a gel using a laser scanner.

Usage:
    laser_scanner <optics>...

<%! from stepwise_mol_bio import hanging_indent %>\
Arguments:
    <optics>
        A laser/filter combination to use.  The most convenient way to specify 
        such a combination is to give the name of a preset.  The following 
        presets are currently available: 

        ${hanging_indent(app.preset_briefs, 8*' ')}

        You can define new presets by adding blocks like the following to your 
        stepwise configuration file:

            [molbio.laser.presets.name]
            laser = <int> or <list of ints>
            filter = <str> or <list of strs>

        You can also explicitly provide laser and filter parameters using the 
        following syntax:

            <laser>/<filter>
"""
    __config__ = [
            DocoptConfig,
            PresetConfig,
            StepwiseConfig.setup('molbio.laser'),
    ]

    presets = appcli.param(
            appcli.Key(StepwiseConfig, 'presets'),
            pick=list,
    )
    preset_briefs = appcli.config_attr()
    preset_brief_template = '{laser} nm'

    # This attribute is a list of:
    # - name of preset (str)
    # - laser/filter (str)
    # - {laser: filter} (dict)
    optics = appcli.param(
            appcli.Key(DocoptConfig, '<optics>'),
            default_factory=list,
    )

    def __init__(self, *optics):
        self.optics = list(optics)

    @classmethod
    def from_laser_filter_pair(cls, laser, filter):
        self = cls()
        self.optics.append({
            'laser': laser,
            'filter': filter,
        })
        return self

    def get_protocol(self):
        optics = [self.parse_optics(x) for x in self.optics]
        lasers = [f"{plural(optics):laser/s}:"] + [f"{x['laser']} nm" for x in optics]
        filters = [f"{plural(optics):filter/s}:"] + [x['filter'] for x in optics]

        p = stepwise.Protocol()
        p += pl(
                "Image with a laser scanner:",
                pre(stepwise.tabulate([lasers, filters])),
        )
        return p

    def parse_optics(self, optics):
        if isinstance(optics, dict):
            return optics

        try:
            return self.presets[optics]
        except ConfigError:
            try: 
                laser, filter = optics.split('/')
                return {'laser': laser, 'filter': filter}
            except ValueError as err:
                raise UsageError(f"expected a preset or '<laser>/<filter>', got {optics!r}") from None

if __name__ == '__main__':
    LaserScanner.main()



#!/usr/bin/env python3

import stepwise, appcli, autoprop
from stepwise import Quantity
from stepwise_mol_bio import Cleanup, UsageError
from freezerbox import MakerConfig, group_by_identity, parse_volume, parse_conc
from appcli import DocoptConfig, Key
from inform import plural

@autoprop
class Lyophilize(Cleanup):
    """\
Concentrate samples by lyophilization.

Usage:
    lyophilize [-v <µL> | -c <conc>]

Options:
    -v --volume <µL>
        The volume to bring the samples to after lyophilization.  If no unit is 
        specified, µL is assumed.

    -c --conc <conc>
        The concentration to bring the samples to after lyophilization.  
        Include a unit.  It is assumed that the experimenter will measure the 
        concentration of the sample before lyophilization and calculate the 
        correct volume of buffer to add.
"""
    __config__ = [
            DocoptConfig,
            MakerConfig,
    ]
    volume = appcli.param(
            Key(DocoptConfig, '--volume', cast=lambda x: parse_volume(x, default_unit='µL')),
            Key(MakerConfig, 'volume', cast=parse_volume),
            default=None,
    )
    conc = appcli.param(
            Key(DocoptConfig, '--conc', cast=parse_conc),
            Key(MakerConfig, 'conc', cast=parse_conc),
            default=None,
    )

    group_by = {
        'volume': group_by_identity,
        'conc': group_by_identity,
    }

    def __init__(self, *, volume=None, conc=None):
        if volume: self.volume = volume
        if conc: self.conc = conc

    def get_protocol(self):
        phrases = []
        n = plural(self.product_tags)

        if self.show_product_tags:
            phrases.append(f"Concentrate the following {n:sample/s}")
        elif self.product_tags:
            phrases.append(f"Concentrate the {n:sample/s}")
        else:
            phrases.append("Concentrate the sample(s)")

        if self.volume and self.conc:
            err = UsageError(volume=self.volume, conc=self.conc)
            err.brief = "cannot specify volume and concentration"
            err.info += "volume: {volume}"
            err.info += "conc: {conc}"
            raise err
        elif self.volume:
            phrases.append(f"to {self.volume}")
        elif self.conc:
            phrases.append(f"to {self.conc}")
        else:
            pass

        if self.show_product_tags:
            phrases.append(f"by lyophilization: {', '.join(map(str, self.product_tags))}")
        else:
            phrases.append("by lyophilization.")

        return stepwise.Protocol(
                steps=[' '.join(phrases)],
        )

    def get_product_conc(self):
        return self.conc

    def get_product_volume(self):
        return self.volume

if __name__ == '__main__':
    Lyophilize.main()
